/* 
 * $Id: bootblock.S,v 1.1.1.1 2007/09/18 14:31:52 ydong Exp $
 * Bootloader for MiniOS:
 * Copy kernel from floppy disk to memory and jump to kernel entry
 * Certified with GCAP1, see PLDI2007, Certified Self-Modifying Code
 * By Hongxu Cai, Zhong Shao and Alexander Vaynberg,
 */
    .equ    BOOT_SEGMENT,  0x07c0 	# equsymbol, expression = .set
    .equ    STACK_SEGMENT, 0x9000	# set symbol value to the expression
    .equ    STACK_BOTTOM,  0xfffe
    .equ    KERNEL_SEGMENT,0x0000	# entry point of kernel code
    .equ    KERNEL_OFFSET, 0x1000
    .equ    KERNELD_SEGMENT,0x0000	# data section segment

.text					# Code segment
.globl  _start				# The entry point must be global
.code16					# Real mode

/*
 * The first instruction to execute in a program is called the entry
 * point. The linker expects to find the entry point in the "symbol" _start
 */
_start:	
        jmp     over
os_size:
        .word   0			# for createimage to write OS size
over:
        movw    $BOOT_SEGMENT,%ax 	# set up bootloader's data segment
        movw    %ax,%ds			# ds = BOOT_SEGMENT (what cs should be)
        movw    $STACK_SEGMENT,%bx	# set up stack
        movw    %bx,%ss			# stack segment
        movw    $STACK_BOTTOM, %sp	# stack pointer
        movw    $KERNEL_SEGMENT, %bx	# load kernel from floppy disk
        movw    %bx,%es			# To para 0: kernel segment
        movw    $KERNEL_OFFSET,%bx	# To para 1: kernel offset
        movw    $os_size,%si		# kernel size
        movb    (%si), %al		# To para 2: sector number
        movb    $2, %ah 		# disk read command
        movb    $0, %ch 		# From para 0: starting cylinder
        movb    $2, %cl 		# From para 1: starting sector
        movb    $0, %dh 		# From para 2: starting head
        int     $0x13			# call INT 13 to read floppy disk
        movw    $KERNELD_SEGMENT,%bx	# set up kernel's data segment
        movw    %bx,%ds			# ds = KERNEL_SEGMENT (the same as cs)
        ljmp    $KERNEL_SEGMENT,$KERNEL_OFFSET # jump to the new kernel entry
forever: 				# loop if the jump failed
        jmp     forever			# should never be here
