/*
 * $Id: util.S,v 1.1.1.1 2007/09/18 14:31:52 ydong Exp $
 * Lib functions for Queue options (enQueue, deQueue) in critial region
 * It's not interruptable
 * Certified with SCAP, see PLDI06 by Xinyu Feng et. al.:
 * Modular Verification of Assembly Code with Stack-Based Control Abstractions.
 */
	.set	Q_HEAD, 0		# field for Queue: head
	.set	Q_TAIL, 2		# field for Queue: tail
	.set	ND_PREV, 0		# field for nodes in Q: prev
	.set	ND_NEXT, 2		# field for nodes in Q: next

.text					# Code section
.code16					# Real mode
.globl enQueue				# called by kernel
.globl deQueue

/* void enQueue(&free_TCB, &free_list)  */
enQueue:				# 0x11e4
	pushw	%bp
	movw	%sp, %bp		# (stack: 0, bp; 2, ra; 4, ax; 6, bx)
	movw	4(%bp), %ax		# ax: pointer to node
	movw	6(%bp), %bx		# bx: pointer to Q
        movw    Q_TAIL(%bx), %dx
	movw	%bx, %cx		# save bx
	movw	%ax, %bx
        movw    %dx, ND_PREV(%bx)
        movw    $0, ND_NEXT(%bx)
        cmpw    $0, %dx
        je      enq_null
        movw    %dx, %bx
        movw    %ax, ND_NEXT(%bx)
	movw	%cx, %bx		# bx: pointers to Q
        movw    %ax, Q_TAIL(%bx)
	popw	%bp
        retw
enq_null:				# 0x120d
	movw	%cx, %bx		# bx: pointers to Q
        movw    %ax, Q_HEAD(%bx)	# head = newtcb
        movw    %ax, Q_TAIL(%bx)
	popw	%bp
        retw

/* *tcb deQueue(&Queue) */
deQueue:
	pushw	%bp
	movw	%sp, %bp		# stack: 0, bp; 2, ra; 4, bx = Q 
	movw	4(%bp), %bx		# bx: pointer to Q
	movw	Q_HEAD(%bx), %dx	# dx: header
        cmpw    $0, %dx			# Q header = NULL, empty Q
        je	no_node			# return NULL
	movw	%bx, %cx		# get a node, save bx
	movw	%dx, %bx		# Q_header
        movw	ND_NEXT(%bx), %dx	# Q_header = Q_header->next
	movw	$0, ND_NEXT(%bx)	# clear return node's next area
	movw	%bx, %ax		# return value
	movw	%cx, %bx		# bx: pointer to Q
        movw    %dx, Q_HEAD(%bx)	# Q_header -> next be new Q header
        cmpw	$0, %dx			# one node, Q be empty after deQ
        je	one_node		# clear Q_Tail
	movw	%dx, %bx		# Q still has some nodes
	movw	$0, ND_PREV(%bx)	# clean header's PREA area
	popw	%bp
	retw				# 1: return TCB, Q still has nodes
no_node:
        movw    $0, %ax			# return value NULL
	popw	%bp
	retw				# 2: return NULL, empty Q
one_node:
        movw    $0, Q_TAIL(%bx)		# Q be empty
	popw	%bp
	retw				# 3: return TCB, empty Q
