/*
 * $Id: init.S,v 1.6 2007/12/09 22:43:15 ydong Exp $
 * miniOS initial functions
 * Initial functions: interrupt handle, memory(free TCBs), threads(idle, echos)
 * precondition for all the functions: interrupt is turned off
 */
	.set	NTIMER_IP, 4*32		# PIC constant for timer interrupt
	.set	NTIMER_CSEG, NTIMER_IP + 2
	.set	PREEMPT_TICS, 11932	# Timer 1193180/11932 = 100 Hz

    	.set    STACK_SZ, 0x400		# 1 kilo stack for user thread
    	.set    STACK_BOTTOM, 0xfffe

	.set	MAX_THRD_NUM, 4		# max number of thread: idle + 3 echoes

    	.set    TCB_SZ, 10		# size of Thread Control Blocks
    	.set    TCB_PREV, 0		# field 1: previous TCB
    	.set    TCB_NEXT, 2		# field 2: next TCB
    	.set    TCB_STACK, 4		# field 3: saved sp
    	.set    TCB_STATUS, 6		# field 4: thread status
	.set	TCB_ID, 8		# field 5: TCB (thread) ID

	.set	DEFAULT_MASK, 0xfe	# interrupt mask allowing only timer

    	.set    READY, 0		# Thread Status

.data                   		# Data section, need memory space, all zero
stack_bot:				# stack bottom for user thread
	.word	0			# stack_bottom = STACK_BOTTOM - STACK_SZ

memory_TCBs: 				# memory for all the TCBs
	.fill	TCB_SZ*MAX_THRD_NUM

free_TCBs:				# Queue for all the free TCBs
	.word 0		# Q_HEAD
	.word 0		# Q_TAIL

readyQ:					# Queue for all the ready thread 
	.word 0		# Q_HEAD	# scheduler will select one to run
	.word 0		# Q_TAIL

curr_thrd:				# points to the current thread
	.word 0

idle_thrd:				# idle thread, not in readyQ
	.word 0				# not be scheduled unless readyQ empty

.text					# Code segment
.code16					# Real mode
.globl	TCB_ID
.globl	TCB_STACK
.globl	readyQ
.globl	curr_thrd
.globl	idle_thrd
.globl	interrupt_init          # initialize timer interrupt handler
.globl	fTCBs_init              # initialize free_TCBs
.globl	idle_init               # initilize idle thread
.globl	spawn                   # spawn MAX_THRD_NUM - 1 echos

/* initialize timer interrupt handler  */
interrupt_init:
        movw    $0x20, %dx              #
        movb    $0x15, %al		# ICW1: 10101b
        outb    %al, %dx                # initialize PIC1
        movw    $0x21, %dx
        movb    $0x20, %al              # ICW2
        outb    %al, %dx                # remap interrupt vector: INT20
        movb    $0x04, %al              # ICW3
        outb    %al, %dx                # PIC2 attached to pin 2
        movb    $0x01, %al              # ICW4
        outb    %al, %dx                # 8086/8088 mode
        movb    $DEFAULT_MASK, %al
        outb    %al, %dx                # mask all except timer and keyboard

        movw    $0xa0, %dx              # initialize PIC2
        movb    $0x15, %al
        outb    %al, %dx                # ICW1: 10101
        movw    $0xa1, %dx
        movb    $0x70, %al              # ICW2: remap vector: 0x70
        outb    %al, %dx
        movb    $0x02, %al              # ICW3: master attached to pin 1
        outb    %al, %dx
        movb    $0x01, %al              # ICW4: 8086/8088 mode
        outb    %al, %dx
        movb    $0xff, %al              # mask all
        outb    %al, %dx

	movw	%ds, %cx		# get and save ds into cx
	movw	%cs, %dx		# get cs into dx
	movw	%dx, %ds		# ds = cs = 0x0000 now
        movw    $PREEMPT_TICS, %ax
        movw    $0x40, %dx              # PIT: programmable interval timer
        outb    %al, %dx
        movb    %ah, %al
        outb    %al, %dx                # set timer interrupt frequency
        movw    $NTIMER_CSEG, %bx
	movw	%ds, %ax
	movw    %ax, (%bx)              # cs = ds for timer interrupt handler
        movw    $NTIMER_IP, %bx
        movw    $timer, (%bx)           # set timer handler
	movw	%cx, %ds		# recover ds, so data section is relocatable now
	retw
	
/* initial all the free TCBs and link them into free_TCBs as a D-link */
fTCBs_init:
	movw	$0, %cx			# %cx is the counter
	movw	$free_TCBs, %bx		
	movw	$memory_TCBs, %ax		
fTCBs_init_loop:			
	cmpw	$MAX_THRD_NUM, %cx	# compare counter with MAX
	je	fTCBs_init_done		# Done, linked all nodes
	pushw	%cx			# save counter on stack
	pushw	%bx			# free_TCBs
	pushw	%ax			# node
	call	enQueue			# enQueue( node, list_head)
	popw	%ax			
	popw	%bx
	popw	%cx			# recover counter
	addw	$TCB_SZ, %ax
	addw	$1, %cx			# counter++
	jmp	fTCBs_init_loop
fTCBs_init_done:			
	retw

/* initilaze idle thread, ID = 0, will scheduled in when readyQ is empty */
idle_init:
        pushw   $free_TCBs
        call    deQueue
        addw    $2, %sp
	movw	%ax, %bx		# got the first TCB for idle
	movw	%ax, (curr_thrd)	# curr_thrd:
	movw	%ax, (idle_thrd)	# idle_thrd:
        movw    $READY, TCB_STATUS(%bx) # initial "idle" thread 0.
	movw	$0, TCB_ID(%bx)		# id of idle thread is ZERO
	retw

/* create, initialize threads and put them into readyQ. Use all the TCBs */
spawn:
        pushw   %bp
        movw    %sp, %bp
	movw	$STACK_BOTTOM, %cx	
	subw	$STACK_SZ, %cx
	movw	%cx, (stack_bot)	# initialize stack bottom
	movw	$1, %cx			# counter, threads ID
spawn_loop:
	call	rgc_cli
	pushw	%cx			# backup cx, it's a counter
        pushw   $free_TCBs		# TCB
        call    deQueue			# get a free TCB (**share resource**)
        addw    $2, %sp
	pop	%cx			# recover cx, thread ID
        movw    (stack_bot), %bx	# Stack: stack_bot for new thread
        cmpw    $0, %ax			# return value ax, pointer of TCB
        je      spawn_done		# Done. all TCBs used
        movw    %bx, %dx		
        subw    $STACK_SZ, %dx		# stack_bot (***share reource***)
        movw    %dx, (stack_bot)	# stack_bot for next thread
	call	rgc_sti
        movw    4(%bp), %dx		# new thread initialization
        pushw   %dx                     # thread entry
        pushw   %ax                     # pointer to TCB
	pushw	%cx			# thread ID
        pushw   %bx                     # stack bottom
        call    init_thrd
        addw    $2, %sp			# reverse stack for reg bx
        popw    %cx			# reverse reg cx
        popw    %ax			# reverse reg ax
        addw    $2, %sp			# reverse stack for reg dx
	addw	$1, %cx			# counter, thread ID ++
        jmp	spawn_loop		# spawn more threads
spawn_done:
	call	rgc_sti
        popw    %bp
        retw

/* initialize stack and TCB, then put it into readyQ */
init_thrd:
        pushw   %bp			# 0,bp; 2,ra; 4,cx; 6,bx; 8,ax; 10,dx
        movw    %sp, %bp
        movw    10(%bp), %ax            # thread entry
        movw    8(%bp), %bx             # bx points to TCB
        movw    6(%bp), %cx             # now cx is thread ID 
        movw    4(%bp), %bp             # bp is the stack top of this thread

        movw    %ax, -6(%bp)            # save entry on the thread stack
        subw    $22, %bp                # reserve space for regs and flags
        movw    $timer_quit, (%bp)	# thread will recover form this point
        movw    %bp, %ax                # ax now contains thread stack top

        movw    $READY, TCB_STATUS(%bx)	# TCB initialization
        movw    %ax, TCB_STACK(%bx)
        movw    %cx, TCB_ID(%bx)

	call	rgc_cli			# readyQ, (** share resource **)
        pushw   $readyQ
        pushw   %bx
        call    enQueue			# link this TCB into readyQ
        addw    $4, %sp
        popw    %bp
	call	rgc_sti
        retw

/* Error, reboot the machine. should never be here*/
reboot:					
	movw	0x472, %bx
	movw 	$0x1234,(%bx)		# Warm boot
        ljmp 	$0xffff, $0x0		# Warm reboot, should never be here
