/*
 * $Id: init.S,v 1.6 2007/12/09 22:43:15 ydong Exp $
 * toy hypervisor (Ranger) initial file
 * Hypervisor, named Ranger; each miniOS, named Cabin.
 * Initial functions: interrupt handle, memory(free CCBs), Cabins(miniOSs)
 * precondition for all the functions: interrupt is turned off
 */
#include "ccb.h"
.data                   		# Data section, need memory space, all zero
cstack_bot:				# stack bottom for cabin
	.word	CSTACK_BOTTOM - CSTACK_SZ

tstack_seg:				# stack segment for each OS
	.word	CSTACK_SEGMENT

tdata_seg:				# data segment for each OS
	.word	TDATA_SEGMENT

memory_CCBs: 				# memory for all the CCBs
	.fill	CCB_SZ*MAX_CABIN_NUM

free_CCBs:				# Queue for all the free CCBs
	.word 0		# Q_HEAD
	.word 0		# Q_TAIL

readyCQ:				# Queue for all the ready cabins 
	.word 0		# Q_HEAD	# scheduler will select one to run
	.word 0		# Q_TAIL

curr_cabin:				# points to the current cabin 
	.word 0

curr_bx:				# points to the current bx
	.word 0				# swap temp var for stack switch

curr_ax:				# points to the current ax
	.word 0				# swap temp var for stack switch

.text					# Code segment
.code16					# Real mode
.globl	readyCQ
.globl	curr_cabin
.globl	curr_bx
.globl	curr_ax
.globl	rg_interrupt_init          	# initialize timer interrupt handler
.globl	rg_fCCBs_init              	# initialize free_CCBs
.globl	rg_run_first                   	# run the first cabin
.globl	rg_build                   	# create MAX_CABIN_NUM-1 cabins

/* initialize timer interrupt handler  */
rg_interrupt_init:
        movw    $0x20, %dx              #
        movb    $0x15, %al		# ICW1: 10101b
        outb    %al, %dx                # initialize PIC1
        movw    $0x21, %dx
        movb    $0x20, %al              # ICW2
        outb    %al, %dx                # remap interrupt vector: INT20
        movb    $0x04, %al              # ICW3
        outb    %al, %dx                # PIC2 attached to pin 2
        movb    $0x01, %al              # ICW4
        outb    %al, %dx                # 8086/8088 mode
        movb    $DEFAULT_MASK, %al
        outb    %al, %dx                # mask all except timer and keyboard

        movw    $0xa0, %dx              # initialize PIC2
        movb    $0x15, %al
        outb    %al, %dx                # ICW1: 10101
        movw    $0xa1, %dx
        movb    $0x70, %al              # ICW2: remap vector: 0x70
        outb    %al, %dx
        movb    $0x02, %al              # ICW3: master attached to pin 1
        outb    %al, %dx
        movb    $0x01, %al              # ICW4: 8086/8088 mode
        outb    %al, %dx
        movb    $0xff, %al              # mask all
        outb    %al, %dx

	movw	%ds, %cx		# get and save ds into cx
	movw	%cs, %dx		# get cs into dx
	movw	%dx, %ds		# ds = cs = 0x0000 now
        movw    $PREEMPT_TICS, %ax
        movw    $0x40, %dx              # PIT: programmable interval timer
        outb    %al, %dx
        movb    %ah, %al
        outb    %al, %dx                # set timer interrupt frequency
        movw    $NTIMER_CSEG, %bx
	movw	%ds, %ax
	movw    %ax, (%bx)              # cs = ds for timer interrupt handler
        movw    $NTIMER_IP, %bx
        movw    $rg_timer, (%bx)        # set timer handler
	movw	%cx, %ds		# recover ds, so data section is relocatable now
	retw
	
/* initial all the free CCBs and link them into free_CCBs as a D-link */
rg_fCCBs_init:
	movw	$0, %cx			# %cx is the counter
	movw	$free_CCBs, %bx		
	movw	$memory_CCBs, %ax		
fCCBs_init_loop:			
	cmpw	$MAX_CABIN_NUM, %cx	# compare counter with MAX
	je	fCCBs_init_done		# Done, linked all nodes
	pushw	%cx			# save counter on stack
	pushw	%bx			# free_TCBs
	pushw	%ax			# node
	call	enQueue			# enQueue( node, list_head)
	popw	%ax			
	popw	%bx
	popw	%cx			# recover counter
	addw	$CCB_SZ, %ax
	addw	$1, %cx			# counter++
	jmp	fCCBs_init_loop
fCCBs_init_done:			
	retw

/* rg_build (_start). build, initialize all the cabins and put them into readyCQ */
rg_build:
        pushw   %bp
        movw    %sp, %bp
	movw	$1, %cx			# counter, cabin ID
build_loop:
	pushw	%cx			# backup cx, it's a counter
        pushw   $free_CCBs		# CCB
        call    deQueue			# get a free CCB (**share resource**)
        addw    $2, %sp
	pop	%cx			# recover cx, thread ID
        movw    (cstack_bot), %bx	# Stack: stack_bot for new thread
        cmpw    $0, %ax			# return value ax, pointer of CCB
        je      build_done		# Done. all CCBs used
        movw    %bx, %dx		
        subw    $CSTACK_SZ, %dx		# stack_bot (***share reource***)
        movw    %dx, (cstack_bot)	# stack_bot for next cabin in ranger

        movw    (tstack_seg), %dx	# stack for cabin's thread
	subw	$TSTACK_SZ, %dx
	movw	%dx, (tstack_seg)
	pushw	%dx			# stack segment for this cabin

        movw    (tdata_seg), %dx	# data for cabin's thread
	addw	$TDATA_SZ, %dx
	movw	%dx, (tdata_seg)
	pushw	%dx			# data segment for this cabin

        movw    4(%bp), %dx		# _start of miniOS
        pushw   %dx                     # thread entry
        pushw   %ax                     # pointer to CCB
	pushw	%cx			# thread ID
        pushw   %bx                     # stack bottom
        call    init_cabin
        addw    $2, %sp			# reverse stack for reg bx
        popw    %cx			# reverse reg cx
        popw    %ax			# reverse reg ax
        addw    $6, %sp			# reverse stack dx: entry, DS, SS
	addw	$1, %cx			# counter, thread ID ++
        jmp	build_loop		# spawn more threads
build_done:
        popw    %bp
        retw

/* initialize stack and CCB, then put it into readyCQ */
init_cabin:
        pushw   %bp			# 0,bp; 2,ra; 4,cx; 6,bx; 8,ax; 10,dx
        movw    %sp, %bp
        movw    14(%bp), %cx            # stack segment for OS
        movw    12(%bp), %bx            # data segment for OS
        movw    10(%bp), %ax            # miniOS entry _start
        movw    4(%bp), %bp             # bp is the stack top of this thread
        movw    %cx, -6(%bp)            # **cabin K: os stack segment
        movw    $rg_timer_quit, -22(%bp)	# cabin will recover form this point
 
	movw	%cx, %ss		# **os K: SS switch 
	movw	$TSTACK_BOTTOM, %bp
        movw    %ax, 8(%bp)             # OS entry on thread stack, next should be timer
        movw    %bx,  (%bp)             # save data segment on thread stack
	movw	$CSTACK_SEGMENT, %cx
	movw	%cx, %ss		# recover SS for ranger

        movw    %sp, %bp		# ** Mem: CCB, rg data segment
        movw    8(%bp), %bx             # bx points to CCB
        movw    6(%bp), %cx             # now cx is thread ID 
        movw    4(%bp), %bp             # bp is the stack top of this thread

        subw    $22, %bp                # reserve space for regs and flags
        movw    %bp, %ax                # ax now contains thread stack top

        movw    $READY, CCB_STATUS(%bx)	# CCB initialization, ready
        movw    $DISAB, CCB_INTFLG(%bx)	# interrupt disable
        movw    %ax, CCB_CSTACK(%bx)	# cabin stack
        movw    $TSTACK_BOTTOM, CCB_TSTACK(%bx)	# os stack
        movw    %cx, CCB_ID(%bx)	# cabin id

        pushw   $readyCQ
        pushw   %bx
        call    enQueue			# link this TCB into readyQ
        addw    $4, %sp
        popw    %bp
        retw


/* run first cabin, ID = 1, just prepare to run and waiting for timer interrupt */
rg_run_first:
        pushw   $readyCQ
        call    deQueue
        addw    $2, %sp
        movw    %ax, %bx                # load_context(deQueue's ra)
        movw    %bx, (curr_cabin)       # resume selected cabin
        movw    CCB_CSTACK(%bx), %sp    # set stack for selected cabin
	addw	$2, %sp			# skip rg_timer_quit
        popw    %bp                     # Cabin's register recover: stack => reg
        popw    %si
        popw    %di
        popw    %dx
        popw    %cx
        popw    %bx
        popw    %ax
        popw    %ss			# stack segment switch: c -> t

        movw    %bx, (curr_bx)          # stack switch: c-sp --> t-sp
        movw    (curr_cabin), %bx
        movw    %sp, CCB_CSTACK(%bx)
        movw    CCB_TSTACK(%bx), %sp
        movw    (curr_bx), %bx          # stack switch: c-sp --> t-sp
	
        popw    %ds			# pop from t-sp, DS: c-ds -> t-ds
	addw	$6, %sp			# skip sp cell for timer iret
	iret				# return to miniOS's first command 

/* Error, reboot the machine. should never be here*/
reboot:					
	movw	0x472, %bx
	movw 	$0x1234,(%bx)		# Warm boot
        ljmp 	$0xffff, $0x0		# Warm reboot, should never be here
