/*
 * $Id: kernel.S,v 1.6 2007/10/05 19:39:28 ydong Exp $
 * miniOS Kernel: Timer(scheduler)
 * See: Certified  Concurrent Code with Interrupts AIM2's timer handler
 * Key for verfification: interrupt + function call, AIM2
 */
#include "ccb.h"
.text					# Code section
.code16					# Real mode
.globl  _start                          # entry poine should be globle
.globl	rg_timer
.globl	rg_timer_quit
/*
 * Entry point: the first instruction to execute
 * Main flow of system initilization: interrupt, memory, threads(idle, echos)
 */
	.set TEMP_EFLAG, 0x0282
# precondition: interrupt is turned off, init critical region in
_start:
	cli
        call    rg_interrupt_init       # initialize ranger timer interrupt handler
        # precondition: sp = 0xfffe, ss = 0x9000
        call    rg_fCCBs_init           # initialize free_CCBs
        pushw   $KERNEL_ENTRY           # parameter of build, entry point of miniOS
        call    rg_build                # build MAX_THRD_NUM cabins
        addw    $2, %sp                 # range bootup finished
	call    rg_run_first            # run first cabin, never return here
idle:					# infinite idle loop, waiting for timer
	jmp	idle			# idle thread will be switch out here

/* Timer interrupt handler, system heart beat, call scheduler here */
# precondition: interrupt is truned off, (pi, gi)
rg_timer:				# h_entry
	pushw	$TEMP_EFLAG
	pushw	%cs
	pushw	$KERNEL_TIMER		# set fake return address to timer
        pushw   %ds			# ra and DS in OS/thread's stack

	pushw	%cx			# ** DS switch: cabin -> ranger
	movw	$CDATA_SEGMENT, %cx
	movw	%cx, %ds
	popw	%cx			# recover cx

	movw	%bx, (curr_bx)		# ** SP switch: t-sp --> c-sp
	movw	(curr_cabin), %bx
        movw	%sp, CCB_TSTACK(%bx)
        movw	CCB_CSTACK(%bx), %sp

	movw	%ss, %bx		# ** SS switch: cabin -> ranger
	movw	%ax, (curr_ax)		# save ax to temp mem cell
	movw	$CSTACK_SEGMENT, %ax	
	movw	%ax, %ss		# updata ss

	pushw	%bx			# push cabin's SS
	movw	(curr_ax), %ax		# recover ax
        pushw   %ax                     
	movw	(curr_bx), %bx		# recover bx
        pushw   %bx
        pushw   %cx
        pushw   %dx
        pushw   %di
        pushw   %si
        pushw   %bp

        movw    $0x20, %dx
        movb    $0x20, %al
        outb    %al, %dx                # send EOI, interrupt ACK

	call	rg_scheduler

rg_timer_quit:				# scheduled thread runs form here
        popw    %bp			# interrupt context load: stack => reg
        popw    %si
        popw    %di
        popw    %dx
        popw    %cx
        popw    %bx
        popw    %ax
	popw	%ss

	movw	%bx, (curr_bx)		# stack switch: c-sp --> t-sp
	movw	(curr_cabin), %bx
        movw	%sp, CCB_CSTACK(%bx)
        movw	CCB_TSTACK(%bx), %sp

	movw	%ax, (curr_ax)		# check int_flag
	movw	CCB_INTFLG(%bx), %ax
	cmpw	$DISAB, %ax			
	je	int_disable		# INTFLG=0, interrup disable
	movw	(curr_bx), %bx		# recover bx
	movw	(curr_ax), %ax		# recover ax
        popw    %ds
        iret				# return to Kernel timer

int_disable:
	movw	(curr_bx), %bx		# recover bx
	movw	(curr_ax), %ax		# recover ax
        popw    %ds
	addw	$6, %sp			# interrupt disabled, can not return to timer
        iret				# return to ra
	
/* RR Scheduler of tinyRG, called by timer interrupt handler  */
#Precondition: interrupt is turn off
rg_scheduler:
        movw    (curr_cabin), %bx	# save_context(curr_cabin)
        movw    %sp, CCB_CSTACK(%bx)	# cabin stack sp --> gs_timer_quit
	pushw	$readyCQ 		# save curr_cabin into readyCQ
        pushw   %bx
        call    enQueue			# enQueue(&curr_cabin, &readyCQ)
        addw    $2, %sp			# recover stack, sp -> &readyCQ
        call    deQueue 		# find next cabin from readyCQ.
        addw    $2, %sp			# recover parameter $readyCQ
	movw	%ax, %bx		# load_context(deQueue's ra)
        movw    %bx, (curr_cabin)	# resume selected cabin
        movw    CCB_CSTACK(%bx), %sp	# set stack for selected cabin
        retw 				# resume curr_cabin' stack and return
